# Copyright 2015 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from elasticsearch import exceptions as es_exc
import mock
import operator
from oslo_serialization import jsonutils
import six
import webob.exc

from searchlight.api.v1 import search
from searchlight.common import exception
from searchlight.common import utils
import searchlight.elasticsearch
import searchlight.gateway
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


def _action_fixture(op_type, data, index=None, doc_type=None, _id=None,
                    **kwargs):
    action = {
        'action': op_type,
        'id': _id,
        'index': index,
        'type': doc_type,
        'data': data,
    }
    if kwargs:
        action.update(kwargs)

    return action


def _image_fixture(op_type, _id=None, index='searchlight-search',
                   doc_type='OS::Glance::Image',
                   data=None, **kwargs):
    image_data = {
        'name': 'image-1',
        'disk_format': 'raw',
    }
    if data is not None:
        image_data.update(data)

    return _action_fixture(op_type, image_data, index, doc_type, _id, **kwargs)


# To avoid repeating these in all the mocks
REPO_SEARCH = 'searchlight.elasticsearch.CatalogSearchRepo.search'
REPO_PLUGINS = 'searchlight.elasticsearch.CatalogSearchRepo.plugins_info'


class TestControllerSearch(test_utils.BaseTestCase):

    def setUp(self):
        super(TestControllerSearch, self).setUp()
        self.search_controller = search.SearchController()

    def test_search_all(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch.object(self.search_controller, 'search',
                               return_value={}) as mock_search:
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10
            self.search_controller.search(
                request, query, index_name, doc_type, offset, limit)
            mock_search.assert_called_once_with(
                request, query, index_name, doc_type, offset, limit)

    def test_search_all_repo(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, return_value={}) as mock_search:
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10
            self.search_controller.search(
                request, query, index_name, doc_type, offset, limit)
            mock_search.assert_called_once_with(
                index_name, doc_type, query, offset,
                limit, ignore_unavailable=True)

    def test_search_forbidden(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, side_effect=exception.Forbidden):
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10

            self.assertRaises(
                webob.exc.HTTPForbidden, self.search_controller.search,
                request, query, index_name, doc_type, offset, limit)

    def test_search_not_found(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, side_effect=exception.NotFound):
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10

            self.assertRaises(
                webob.exc.HTTPNotFound, self.search_controller.search, request,
                query, index_name, doc_type, offset, limit)

    def test_search_duplicate(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, side_effect=exception.Duplicate):
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10

            self.assertRaises(
                webob.exc.HTTPConflict, self.search_controller.search, request,
                query, index_name, doc_type, offset, limit)

    def test_search_badrequest(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, side_effect=es_exc.RequestError):
            query = {}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10

            self.assertRaises(
                webob.exc.HTTPBadRequest, self.search_controller.search,
                request, query, index_name, doc_type, offset, limit)

    def test_search_internal_server_error(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, side_effect=Exception):
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10

            self.assertRaises(
                webob.exc.HTTPInternalServerError,
                self.search_controller.search,
                request, query, index_name, doc_type, offset, limit)

    @mock.patch(REPO_SEARCH, return_value={})
    def test_search_policy_check(self, _):
        request = unit_test_utils.get_fake_request()
        with mock.patch.object(self.search_controller.policy,
                               'enforce') as mock_enforce:

            self.search_controller.search(request, query={"match_all": {}})
            mock_enforce.assert_called_with(request.context,
                                            'search:query',
                                            request.context.policy_target)

    def test_search_version(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, return_value={}) as mock_search:
            query = {"match_all": {}}
            index_name = "searchlight"
            doc_type = "OS::Glance::Metadef"
            offset = 0
            limit = 10
            self.search_controller.search(
                request, query, index_name, doc_type, offset, limit,
                version=True)
            mock_search.assert_called_once_with(
                index_name, doc_type, query, offset,
                limit, ignore_unavailable=True, version=True)

    def test_search_aggregations(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_SEARCH, return_value={}) as mock_search:
            query = {"query": {"match_all": {}},
                     "aggs": {"test": {"terms": {"field": "some_field"}}}}
            index_name = "searchlight"
            limit = 10
            offset = 0
            doc_type = "OS::Glance::Image"

            self.search_controller.search(
                request, query, index_name, doc_type, offset, limit,
                version=True)

            mock_search.assert_called_once_with(
                index_name, doc_type, query, offset,
                limit, ignore_unavailable=True, version=True)


class TestControllerPluginsInfo(test_utils.BaseTestCase):

    def setUp(self):
        super(TestControllerPluginsInfo, self).setUp()
        self.search_controller = search.SearchController()

    def test_plugins_info_forbidden(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_PLUGINS, side_effect=exception.Forbidden):
            self.assertRaises(
                webob.exc.HTTPForbidden, self.search_controller.plugins_info,
                request, [])

    def test_plugins_info_not_found(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_PLUGINS, side_effect=exception.NotFound):
            self.assertRaises(webob.exc.HTTPNotFound,
                              self.search_controller.plugins_info, request, [])

    def test_plugins_info_internal_server_error(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch(REPO_PLUGINS, side_effect=Exception):
            self.assertRaises(webob.exc.HTTPInternalServerError,
                              self.search_controller.plugins_info, request, [])

    def test_plugins_info(self):
        request = unit_test_utils.get_fake_request()
        expected = {
            "plugins": [
                {
                    "alias-searching": "searchlight-search",
                    "alias-indexing": "searchlight-listener",
                    "type": "OS::Designate::RecordSet",
                },
                {
                    "alias-searching": "searchlight-search",
                    "alias-indexing": "searchlight-listener",
                    "type": "OS::Designate::Zone",
                },
                {
                    "alias-searching": "searchlight-search",
                    "alias-indexing": "searchlight-listener",
                    "type": "OS::Glance::Image",
                },
                {
                    "alias-searching": "searchlight-search",
                    "alias-indexing": "searchlight-listener",
                    "type": "OS::Glance::Metadef",
                },
                {
                    "alias-searching": "searchlight-search",
                    "alias-indexing": "searchlight-listener",
                    "type": "OS::Nova::Server",
                }
            ]
        }

        # Simulate policy filtering
        doc_types = [p['type'] for p in expected['plugins']]
        actual = self.search_controller.plugins_info(request, doc_types)
        self.assertEqual(['plugins'], list(actual.keys()))

        self.assertEqual(
            sorted(expected['plugins'], key=operator.itemgetter('type')),
            sorted(actual['plugins'], key=operator.itemgetter('type')))


class TestControllerFacets(test_utils.BaseTestCase):
    def setUp(self):
        super(TestControllerFacets, self).setUp()
        self.search_controller = search.SearchController()

    def test_facets(self):
        request = unit_test_utils.get_fake_request()
        doc_types = ["OS::Glance::Image", "OS::Nova::Server"]
        gf_path = 'searchlight.elasticsearch.plugins.base.IndexBase.get_facets'
        with mock.patch(gf_path) as mock_facets:
            mock_facets.return_value = [{"name": "fake", "type": "string"}], 0

            default_response = self.search_controller.facets(
                request, doc_type=doc_types)

            expected = {
                "OS::Nova::Server": {
                    "doc_count": 0,
                    "facets": [{"name": "fake", "type": "string"}]
                },
                "OS::Glance::Image": {
                    "doc_count": 0,
                    "facets": [{"name": "fake", "type": "string"}]
                }
            }
            self.assertEqual(expected, default_response)

            totals_only_response = self.search_controller.facets(
                request, doc_type=doc_types, include_fields=False)

            expected = {"OS::Nova::Server": {"doc_count": 0},
                        "OS::Glance::Image": {"doc_count": 0}}
            self.assertEqual(expected, totals_only_response)


class TestSearchDeserializer(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSearchDeserializer, self).setUp()
        self.deserializer = search.RequestDeserializer(
            utils.get_search_plugins()
        )

    def test_single_index(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'index': 'searchlight-search',
        }))

        output = self.deserializer.search(request)
        self.assertEqual(['searchlight-search'], output['index'])

    def test_single_type(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': 'OS::Glance::Image',
        }))

        output = self.deserializer.search(request)
        self.assertEqual(['searchlight-search'], output['index'])
        self.assertEqual(['OS::Glance::Image'], output['doc_type'])

    def test_empty_request(self):
        """Tests that ALL registered resource types are searched"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({}))

        output = self.deserializer.search(request)
        self.assertEqual(['searchlight-search'], output['index'])

        types = [
            'OS::Designate::RecordSet',
            'OS::Designate::Zone',
            'OS::Glance::Image',
            'OS::Glance::Metadef',
            'OS::Ironic::Chassis',
            'OS::Ironic::Node',
            'OS::Ironic::Port',
            'OS::Nova::Flavor',
            'OS::Nova::Server',
            'OS::Nova::ServerGroup',
            'OS::Nova::Hypervisor',
            'OS::Neutron::FloatingIP',
            'OS::Neutron::Net',
            'OS::Neutron::Port',
            'OS::Neutron::Subnet',
            'OS::Neutron::Router',
            'OS::Neutron::SecurityGroup',
            'OS::Cinder::Volume',
            'OS::Cinder::Snapshot'
        ]

        self.assertEqual(['searchlight-search'], output['index'])
        self.assertEqual(sorted(types), sorted(output['doc_type']))

    def test_forbidden_schema(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'schema': {},
        }))

        self.assertRaises(webob.exc.HTTPForbidden, self.deserializer.search,
                          request)

    def test_forbidden_self(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'self': {},
        }))

        self.assertRaises(webob.exc.HTTPForbidden, self.deserializer.search,
                          request)

    def test_fields_restriction(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': ['description'],
        }))

        output = self.deserializer.search(request)
        self.assertEqual(['searchlight-search'], output['index'])
        self.assertEqual(['OS::Glance::Metadef'], output['doc_type'])
        self.assertEqual(['description'], output['_source_include'])

    def test_fields_include_exclude(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': {
                'include': ['some', 'thing.*'],
                'exclude': ['other.*', 'thing']
            }
        }))

        output = self.deserializer.search(request)
        self.assertNotIn('_source', output)
        self.assertEqual(['some', 'thing.*'], output['_source_include'])
        # Don't test the role filter exclusion here
        self.assertLessEqual(set(['other.*', 'thing']),
                             set(output['_source_exclude']))

    def test_fields_exclude_rbac(self):
        """Test various forms for source_exclude"""
        role_field = searchlight.elasticsearch.ROLE_USER_FIELD
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': {
                'exclude': ['something', 'other thing']
            }
        }))
        output = self.deserializer.search(request)
        self.assertEqual([role_field, 'something', 'other thing'],
                         output['_source_exclude'])

        # Test with a single field
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': {
                'exclude': "something"
            }
        }))
        output = self.deserializer.search(request)
        self.assertEqual([role_field, 'something'],
                         output['_source_exclude'])

        # Test with a single field
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': "includeme"
        }))
        output = self.deserializer.search(request)
        self.assertEqual([role_field],
                         output['_source_exclude'])
        self.assertEqual("includeme",
                         output['_source_include'])

        # Test with a single field
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': ["includeme", "andme"]
        }))
        output = self.deserializer.search(request)
        self.assertEqual([role_field],
                         output['_source_exclude'])
        self.assertEqual(["includeme", "andme"],
                         output['_source_include'])

    def test_bad_field_include(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            '_source': 1234,
        }))

        self.assertRaisesRegex(
            webob.exc.HTTPBadRequest,
            "'_source' must be a string, dict or list",
            self.deserializer.search,
            request)

    def test_highlight_fields(self):
        """Test that if no highlight_query is given for a field the query is
        applied, and that highlight terms make it through the deserializer.
        """
        request = unit_test_utils.get_fake_request()

        highlight_query = {'query_string': {'query': 'gimme everything'}}

        # Apply highlighting to 'name' explicitly setting require_field_match
        # and 'content' explicitly setting a highlight_query
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'highlight': {
                'fields': {
                    'name': {'require_field_match': True},
                    'content': {
                        'highlight_query': highlight_query
                    }
                }
            }
        }))

        output = self.deserializer.search(request)
        self.assertEqual(['searchlight-search'], output['index'])
        self.assertEqual(['OS::Glance::Metadef'], output['doc_type'])

        expected_highlight = {
            'fields': {
                'name': {
                    # Expects the match_all query we passed in, but preserve
                    # require_field_match
                    'highlight_query': {'match_all': {}},
                    'require_field_match': True
                },
                'content': {
                    # Expect the overridden highlight_query and default
                    # require_field_match
                    'highlight_query': highlight_query,
                    'require_field_match': False
                }
            }
        }
        self.assertEqual(expected_highlight, output['query']['highlight'])

    def test_invalid_limit(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'limit': 'invalid',
        }))

        self.assertRaises(webob.exc.HTTPBadRequest, self.deserializer.search,
                          request)

    def test_negative_limit(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'limit': -1,
        }))

        self.assertRaises(webob.exc.HTTPBadRequest, self.deserializer.search,
                          request)

    def test_invalid_offset(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'offset': 'invalid',
        }))

        self.assertRaises(webob.exc.HTTPBadRequest, self.deserializer.search,
                          request)

    def test_negative_offset(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'offset': -1,
        }))

        self.assertRaises(webob.exc.HTTPBadRequest, self.deserializer.search,
                          request)

    def test_offset_from_error(self):
        """Test that providing offset and from cause errors"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'offset': 10,
            'from': 10
        }))
        self.assertRaisesRegex(
            webob.exc.HTTPBadRequest,
            "Provide 'offset' or 'from', but not both",
            self.deserializer.search, request)

    def test_limit_size_error(self):
        """Test that providing limit and size cause errors"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'size': 10,
            'limit': 10
        }))
        self.assertRaisesRegex(
            webob.exc.HTTPBadRequest,
            "Provide 'limit' or 'size', but not both",
            self.deserializer.search, request)

    def test_limit_and_offset(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'limit': 1,
            'offset': 2,
        }))

        output = self.deserializer.search(request)
        self.assertEqual(1, output['size'])
        self.assertEqual(2, output['from_'])

    def test_from_and_size(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'size': 1,
            'from': 2,
        }))
        output = self.deserializer.search(request)
        self.assertEqual(1, output['size'])
        self.assertEqual(2, output['from_'])

    def test_single_sort(self):
        """Test that a single sort field is correctly transformed"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'sort': 'status'
        }))

        output = self.deserializer.search(request)
        self.assertEqual(['status'], output['query']['sort'])

    def test_single_sort_dir(self):
        """Test that a single sort field & dir is correctly transformed"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'sort': {'status': 'desc'}
        }))

        output = self.deserializer.search(request)
        self.assertEqual([{'status': 'desc'}], output['query']['sort'])

    def test_multiple_sort(self):
        """Test multiple sort fields"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'sort': [
                'status',
                {'created_at': 'desc'},
                {'members': {'order': 'asc', 'mode': 'max'}}
            ]
        }))

        output = self.deserializer.search(request)
        expected = [
            'status',
            {'created_at': 'desc'},
            {'members': {'order': 'asc', 'mode': 'max'}}
        ]
        self.assertEqual(expected, output['query']['sort'])

    def test_raw_field_sort(self):
        """Some fields (like name) are treated separately"""
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'sort': [
                'name',
                {'name': {'order': 'desc'}}
            ]
        }))

        output = self.deserializer.search(request)
        expected = [
            'name.raw',
            {'name.raw': {'order': 'desc'}}
        ]
        self.assertEqual(expected, output['query']['sort'])

    def test_bad_sort(self):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'index': ['glance'],
            'type': ['OS::Glance::Image'],
            'query': {'match_all': {}},
            'sort': 1234
        }))

        self.assertRaises(webob.exc.HTTPBadRequest, self.deserializer.search,
                          request)

    @mock.patch('searchlight.elasticsearch.plugins.nova.servers.' +
                'ServerIndex.get_query_filters')
    def test_rbac_exception(self, mock_query_filters):
        request = unit_test_utils.get_fake_request()
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
        }))

        mock_query_filters.side_effect = Exception("Bad RBAC")

        self.assertRaisesRegex(
            webob.exc.HTTPInternalServerError,
            "Error processing OS::Nova::Server RBAC filter",
            self.deserializer.search,
            request)

    def test_rbac_non_admin(self):
        """Test that a non-admin request results in an RBACed query"""
        request = unit_test_utils.get_fake_request(is_admin=False)
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'type': 'OS::Nova::Server',
        }))
        output = self.deserializer.search(request)
        tenant_id = '6838eb7b-6ded-dead-beef-b344c77fe8df'

        nova_rbac_filter = {
            'indices': {
                'query': {
                    'bool': {
                        'filter': {
                            'bool': {
                                'must': {
                                    'type': {'value': 'OS::Nova::Server'}
                                },
                                'should': [
                                    {'term': {'tenant_id': tenant_id}}
                                ],
                                'minimum_should_match': 1
                            }
                        }
                    }
                },
                'index': 'searchlight-search',
                'no_match_query': 'none'
            }
        }

        role_field = searchlight.elasticsearch.ROLE_USER_FIELD
        expected_query = {
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {role_field: 'user'}},
                            'should': [nova_rbac_filter],
                            'minimum_should_match': 1
                        }
                    },
                    'must': {'match_all': {}}
                }
            }
        }
        self.assertEqual(expected_query, output['query'])

    def test_rbac_admin(self):
        """Test that admins have RBAC applied"""
        request = unit_test_utils.get_fake_request(is_admin=True)
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'type': 'OS::Nova::Server',
        }))
        output = self.deserializer.search(request)
        tenant_id = '6838eb7b-6ded-dead-beef-b344c77fe8df'
        nova_rbac_filter = {
            'indices': {
                'query': {
                    'bool': {
                        'filter': {
                            'bool': {
                                'must': {
                                    'type': {'value': 'OS::Nova::Server'}
                                },
                                'should': [
                                    {'term': {'tenant_id': tenant_id}}
                                ],
                                'minimum_should_match': 1
                            }
                        }
                    }
                },
                'index': 'searchlight-search',
                'no_match_query': 'none'
            }
        }

        role_field = searchlight.elasticsearch.ROLE_USER_FIELD
        expected_query = {
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {role_field: 'admin'}},
                            'should': [nova_rbac_filter],
                            'minimum_should_match': 1
                        }
                    },
                    'must': {'match_all': {}}
                }
            }
        }

        self.assertEqual(expected_query, output['query'])

        # Now test with all_projects
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'type': 'OS::Nova::Server',
            'all_projects': True,
        }))

        # Test that if a plugin doesn't allow RBAC to be ignored,
        # it isn't. Do it with mocking, because mocking is best
        with mock.patch('searchlight.elasticsearch.plugins.nova.servers.'
                        'ServerIndex.allow_admin_ignore_rbac',
                        new_callable=mock.PropertyMock) as ignore_mock:
            ignore_mock.return_value = False
            output = self.deserializer.search(request)
            self.assertEqual(expected_query, output['query'])

        # Now test the same under the default allow-ignore-rbac conditions,
        # and we shouldn't see the tenant restrictions
        output = self.deserializer.search(request)

        # No more tenant restriction in the expected result
        nova_index_filter = nova_rbac_filter['indices']['query']['bool']
        del nova_index_filter['filter']['bool']['should']
        del nova_index_filter['filter']['bool']['minimum_should_match']

        self.assertEqual(expected_query, output['query'])

    def test_default_facet_options(self):
        request = unit_test_utils.get_fake_request(path='/v1/search/facets')
        output = self.deserializer.facets(request)

        output['doc_type'] = sorted(output['doc_type'])
        expected_doc_types = sorted(utils.get_search_plugins().keys())
        expected = {'index_name': None, 'doc_type': expected_doc_types,
                    'all_projects': False, 'limit_terms': 0,
                    'include_fields': True, 'exclude_options': False}
        self.assertEqual(expected, output)

    def test_facet_exclude_options(self):
        path = '/v1/search/facets?exclude_options=True'
        request = unit_test_utils.get_fake_request(path=path)
        output = self.deserializer.facets(request)

        output['doc_type'] = sorted(output['doc_type'])
        expected_doc_types = sorted(utils.get_search_plugins().keys())
        expected = {'index_name': None, 'doc_type': expected_doc_types,
                    'all_projects': False, 'limit_terms': 0,
                    'include_fields': True, 'exclude_options': True}
        self.assertEqual(expected, output)

    def test_search_version(self):
        request = unit_test_utils.get_fake_request(is_admin=True)
        request.body = six.b(jsonutils.dumps({
            'query': {'match_all': {}},
            'version': True
        }))
        output = self.deserializer.search(request)
        self.assertEqual(True, output['version'])

    def test_search_aggregations(self):
        request = unit_test_utils.get_fake_request()

        aggs = {'names': {'terms': {'field': 'name'}},
                'max_something': {'max': {'field': 'something'}}}

        # Apply highlighting to 'name' explicitly setting require_field_match
        # and 'content' explicitly setting a highlight_query
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'aggregations': aggs}))

        output = self.deserializer.search(request)
        self.assertEqual(aggs, output['query']['aggregations'])

        # Test 'aggs' too
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'aggs': aggs}))
        output = self.deserializer.search(request)
        self.assertEqual(aggs, output['query']['aggregations'])

    def test_global_aggregation_not_allowed(self):
        """The 'global' aggregation type bypasses the search query (thus
        bypassing RBAC). This is Bad, so we won't allow it. It can only occur
        as a top level aggregation.
        """
        request = unit_test_utils.get_fake_request()

        aggs = {
            'cheating_rbac': {
                'global': {},
                'aggregations': {'name': {'terms': {'field': 'name'}}}
            }
        }
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'aggregations': aggs}))

        self.assertRaises(
            webob.exc.HTTPForbidden, self.deserializer.search,
            request)

    def test_search_aggregations_bad_request(self):
        """Test that 'aggs' AND 'aggregations' can't be specified together"""
        request = unit_test_utils.get_fake_request()

        aggs = {"something": "something"}
        request.body = six.b(jsonutils.dumps({
            'type': ['OS::Glance::Metadef'],
            'query': {'match_all': {}},
            'aggregations': aggs,
            'aggs': aggs}))

        self.assertRaisesRegex(
            webob.exc.HTTPBadRequest,
            "A request cannot include both 'aggs' and 'aggregations'",
            self.deserializer.search, request)


class TestResponseSerializer(test_utils.BaseTestCase):
    def setUp(self):
        super(TestResponseSerializer, self).setUp()
        self.serializer = search.ResponseSerializer()

    def test_plugins_info(self):
        expected = {
            'plugins': [
                {
                    'OS::Glance::Image': {
                        'index': 'searchlight-search',
                        'type': 'OS::Glance::Image'
                    }
                },
                {
                    'OS::Glance::Metadef': {
                        'index': 'searchlight-search',
                        'type': 'OS::Glance::Metadef'
                    }
                }
            ]
        }

        request = webob.Request.blank('/v0.1/search')
        response = webob.Response(request=request)
        result = {
            'plugins': [
                {
                    'OS::Glance::Image': {
                        'index': 'searchlight-search',
                        'type': 'OS::Glance::Image'
                    }
                },
                {
                    'OS::Glance::Metadef': {
                        'index': 'searchlight-search',
                        'type': 'OS::Glance::Metadef'
                    }
                }
            ]
        }
        self.serializer.search(response, result)
        actual = jsonutils.loads(response.body)
        self.assertEqual(expected, actual)
        self.assertEqual('application/json', response.content_type)

    def test_search(self):
        expected = [{
            'id': '1',
            'name': 'image-1',
            'disk_format': 'raw',
        }]

        request = webob.Request.blank('/v0.1/search')
        response = webob.Response(request=request)
        result = [{
            'id': '1',
            'name': 'image-1',
            'disk_format': 'raw',
        }]
        self.serializer.search(response, result)
        actual = jsonutils.loads(response.body)
        self.assertEqual(expected, actual)
        self.assertEqual('application/json', response.content_type)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.nova import FLAVOR_ACCESS_FIELD
from searchlight.elasticsearch.plugins.nova \
    import notification_handler
from searchlight.elasticsearch.plugins.nova import serialize_nova_flavor
from searchlight.elasticsearch.plugins import openstack_clients


class FlavorIndex(base.IndexBase):

    NotificationHandlerCls = notification_handler.FlavorHandler

    @classmethod
    def get_document_type(self):
        return resource_types.NOVA_FLAVOR

    def get_mapping(self):
        string_not_analyzed = {'type': 'string', 'index': 'not_analyzed'}
        integer = {'type': 'integer'}
        return {
            'dynamic': True,
            'properties': {
                'id': string_not_analyzed,
                'tenant_id': string_not_analyzed,
                'OS-FLV-DISABLED:disabled': {'type': 'boolean'},
                'OS-FLV-EXT-DATA:ephemeral': integer,
                'disk': integer,
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': string_not_analyzed
                    }
                },
                'os-flavor-access:is_public': {'type': 'boolean'},
                FLAVOR_ACCESS_FIELD: string_not_analyzed,
                'ram': integer,
                'rxtx_factor': {'type': 'float'},
                'swap': string_not_analyzed,
                'vcpus': integer,
                'extra-specs': {
                    'type': 'object',
                    'properties': {}
                }
            }
        }

    def get_objects(self):
        """Generator that lists all nova Flavors"""
        return openstack_clients.get_novaclient().flavors.list(is_public=None)

    @property
    def facets_with_options(self):
        return ('OS-FLV-DISABLED:disabled', 'os-flavor-access:is_public')

    @property
    def resource_allowed_policy_target(self):
        return 'os_compute_api:flavors'

    @property
    def service_type(self):
        return 'compute'

    def serialize(self, flavor):
        return serialize_nova_flavor(flavor)

    def _get_rbac_field_filters(self, request_context):
        """Return any RBAC field filters to be injected into an indices
        query. Document type will be added to this list.
        """
        return [
            {'term': {'os-flavor-access:is_public': True}},
            {'term': {FLAVOR_ACCESS_FIELD: request_context.tenant}}
        ]

    def filter_result(self, hit, request_context):
        super(FlavorIndex, self).filter_result(hit, request_context)
        # Only admins and tenants who the flavor has been granted
        # access can see the full list of access.
        if not request_context.is_admin:
            source = hit['_source']
            is_public = source.get('os-flavor-access:is_public', False)
            access = source.get(FLAVOR_ACCESS_FIELD, [])
            if is_public or request_context.tenant not in access:
                source.pop(FLAVOR_ACCESS_FIELD, None)



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.neutron import notification_handlers
from searchlight.elasticsearch.plugins.neutron import serialize_security_group
from searchlight.elasticsearch.plugins import openstack_clients


class SecurityGroupIndex(base.IndexBase):
    NotificationHandlerCls = notification_handlers.SecurityGroupHandler

    @classmethod
    def get_document_type(self):
        return resource_types.NEUTRON_SECURITY_GROUP

    def get_mapping(self):
        string_not_analyzed = {'type': 'string', 'index': 'not_analyzed'}
        ordered_string = {'type': 'string',
                          'fields': {'raw': string_not_analyzed}}

        return {
            'dynamic': False,
            'properties': {
                'security_group_rules': {
                    'type': 'nested',
                    'properties': {
                        'remote_group_id': string_not_analyzed,
                        'direction': {'type': 'string'},
                        'protocol': {'type': 'string'},
                        'description': {'type': 'string'},
                        'ethertype': {'type': 'string'},
                        'remote_ip_prefix': {'type': 'string'},
                        'port_range_max': {'type': 'integer'},
                        'port_range_min': {'type': 'integer'},
                        'security_group_id': string_not_analyzed,
                        'tenant_id': string_not_analyzed,
                        'id': string_not_analyzed,
                    },
                },
                'id': string_not_analyzed,
                'project_id': string_not_analyzed,
                'tenant_id': string_not_analyzed,
                'name': ordered_string,
                'description': {'type': 'string'},
            },
            "_meta": {
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
            }
        }

    @property
    def facets_with_options(self):
        return ('security_group_rules.direction',
                'security_group_rules.ethertype')

    @property
    def facets_excluded(self):
        return {'project_id': True, 'tenant_id': True}

    @property
    def resource_allowed_policy_target(self):
        # Neutron only supports policy for individual resources
        return None

    @property
    def service_type(self):
        return 'network'

    def _get_rbac_field_filters(self, request_context):
        return [
            {'term': {'tenant_id': request_context.owner}}
        ]

    def get_objects(self):
        """Generator that lists all security groups."""
        neutron_client = openstack_clients.get_neutronclient()
        for group in neutron_client.list_security_groups()['security_groups']:
            yield group

    def serialize(self, sec_group):
        return serialize_security_group(sec_group)



# Copyright 2015 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from searchlight.common import resource_types
from searchlight.common import utils
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.glance \
    import metadefs_notification_handler
from searchlight.elasticsearch.plugins.glance \
    import serialize_glance_metadef_ns


class MetadefIndex(base.IndexBase):
    NotificationHandlerCls = metadefs_notification_handler.MetadefHandler

    @classmethod
    def get_document_type(cls):
        return resource_types.GLANCE_METADEF

    def get_document_id_field(self):
        return 'namespace'

    def get_mapping(self):
        property_mapping = {
            'dynamic': True,
            'type': 'nested',
            'properties': {
                'name': {'type': 'string', 'index': 'not_analyzed'},
                'type': {'type': 'string'},
                'title': {'type': 'string'},
                'description': {'type': 'string'},
            }
        }
        mapping = {
            'properties': {
                'created_at': {'type': 'date'},
                'updated_at': {'type': 'date'},
                'display_name': {'type': 'string'},
                'description': {'type': 'string'},
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'namespace': {'type': 'string', 'index': 'not_analyzed'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'owner': {'type': 'string', 'index': 'not_analyzed'},
                'project_id': {'type': 'string', 'index': 'not_analyzed'},
                'visibility': {'type': 'string', 'index': 'not_analyzed'},
                'resource_types': {
                    'type': 'object',
                    'properties': {
                        # sjmc7 - removing prefix, properties_target since they
                        # aren't accessible though the API
                        'name': {'type': 'string'}
                    },
                },
                'objects': {
                    'type': 'nested',
                    'properties': {
                        'id': {'type': 'string', 'index': 'not_analyzed'},
                        'name': {'type': 'string'},
                        'description': {'type': 'string'},
                        'properties': property_mapping,
                    }
                },
                'properties': property_mapping,
                'tags': {'type': 'string'}
            },
            "_meta": {
                "owner": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                }
            }

        }
        return mapping

    @property
    def resource_allowed_policy_target(self):
        return 'get_metadef_namespaces'

    @property
    def service_type(self):
        return 'image'

    def _get_rbac_field_filters(self, request_context):
        return [
            {'term': {'owner': request_context.owner}},
            {'term': {'visibility': 'public'}}
        ]

    def get_objects(self):
        from searchlight.elasticsearch.plugins import openstack_clients
        gc = openstack_clients.get_glanceclient()
        return list(gc.metadefs_namespace.list())

    def serialize(self, metadef_obj):
        return serialize_glance_metadef_ns(metadef_obj)

    def filter_result(self, hit, request_context):
        super(MetadefIndex, self).filter_result(hit, request_context)

        # Revert the change we make to fit the 'tags' mapping used in other
        # plugins (see serialize_glance_metadef_ns in __init__.py)
        highlight = hit.get('highlight')
        if highlight:
            utils.restore_dots_in_field_names(highlight)
        source = hit['_source']
        utils.restore_dots_in_field_names(source)
        tags = source.pop('tags', None)
        if tags is not None:
            source['tags'] = [{"name": tag} for tag in tags]



# Copyright 2016 Hewlett-Packard Enterprise Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from searchlight.elasticsearch.plugins import utils


def add_rbac(network, target_tenant, policy_id):
    """Update a network based on an RBAC policy.
    """
    # Add target_tenant to members list.
    members_list = network['members']
    if target_tenant not in members_list:
        members_list.append(target_tenant)

    # Add RBAC policy.
    rbac_policy = network['rbac_policy']
    policy = {'rbac_id': policy_id, 'target_tenant': target_tenant}
    rbac_policy.append(policy)
    network['rbac_policy'] = rbac_policy


def serialize_network(network):
    serialized = copy.deepcopy(network)
    # Remove subnets because we index them separately
    serialized.pop('subnets')
    serialized['project_id'] = serialized['tenant_id']
    if 'members' not in serialized:
        serialized['members'] = []
    if 'rbac_policy' not in serialized:
        serialized['rbac_policy'] = []
    return serialized


def serialize_port(port):
    serialized = copy.deepcopy(port)
    serialized['project_id'] = serialized['tenant_id']
    return serialized


def serialize_subnet(subnet):
    serialized = copy.deepcopy(subnet)
    serialized['project_id'] = serialized['tenant_id']
    return serialized


def serialize_router(router, updated_at=None):
    serialized = copy.deepcopy(router)
    if 'updated_at' not in router:
        serialized['updated_at'] = updated_at or utils.get_now_str()
    serialized['project_id'] = serialized['tenant_id']
    return serialized


def serialize_floatingip(fip, updated_at=None):
    serialized = copy.deepcopy(fip)
    if 'updated_at' not in fip:
        serialized['updated_at'] = updated_at or utils.get_now_str()
    serialized['project_id'] = serialized['tenant_id']
    return serialized


def serialize_security_group(sec_group, updated_at=None):
    serialized = copy.deepcopy(sec_group)
    if 'updated_at' not in sec_group:
        serialized['updated_at'] = updated_at or utils.get_now_str()
    serialized['project_id'] = serialized['tenant_id']
    return serialized



# Copyright 2015 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock
import novaclient.exceptions
import novaclient.v2.servers as novaclient_servers

from searchlight.elasticsearch.plugins.nova import\
    servers as servers_plugin
from searchlight.elasticsearch import ROLE_USER_FIELD
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


USER1 = u'27f4d76b-be62-4e4e-aa33bb11cc55'

TENANT1 = u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0'
TENANT2 = u'c6993374-7c4b-4f18-b317-85e3acdfd259'

# Instances
ID1 = u'6c41b4d1-f0fa-42d6-9d8d-e3b99695aa69'
ID2 = u'08ca6c43-eea8-48d0-bbb2-30c50109d5d8'
ID3 = u'a380287d-1f61-4887-959c-8c5ab8f75f8f'


flavor1 = {
    u'id': '1',
    u'links': [{
        u'href': u'http://localhost:8774/dontcare',
        u'rel': u'bookmark'
    }]
}
flavor2 = {
    u'id': '2',
    u'links': [{
        u'href': u'http://localhost:8774/stilldontcare',
        u'rel': u'bookmark'
    }]
}

imagea = {
    u'id': u'a',
    u'links': [{
        u'href': u'http://localhost:8774/dontcare',
        u'rel': u'bookmark'
    }]
}
imageb = {
    u'id': u'b',
    u'links': [{
        u'href': u'http://localhost:8774/dontcare',
        u'rel': u'bookmark'
    }]
}

net_ip4_6 = {
    u'net4_6': [
        {
            u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
            u'version': 4,
            u'addr': u'0.0.0.0',
            u'OS-EXT-IPS:type': u'fixed'
        },
        {
            u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
            u'version': 6,
            u'addr': u'::1',
            u'OS-EXT-IPS:type': u'fixed'
        }
    ],
    u'net4': [
        {
            u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
            u'version': 4,
            u'addr': u'127.0.0.1',
            u'OS-EXT-IPS:type': u'fixed'
        }
    ]
}

fake_version_list = [test_utils.FakeVersion('2.1'),
                     test_utils.FakeVersion('2.1')]

net_ipv4 = {u'net4': [dict(net_ip4_6[u'net4'][0])]}

_now = datetime.datetime.utcnow()
_five_minutes_ago = _now - datetime.timedelta(minutes=5)
created_now = _five_minutes_ago.strftime('%Y-%m-%dT%H:%M:%SZ')
updated_now = _now.strftime('%Y-%m-%dT%H:%M:%SZ')

nova_server_getter = 'novaclient.v2.client.servers.ServerManager.get'
nova_version_getter = 'novaclient.v2.client.versions.VersionManager.list'


def _instance_fixture(instance_id, name, tenant_id, **kwargs):
    # A full nova v2 server.get output
    attrs = {
        u'OS-DCF:diskConfig': u'MANUAL',
        u'OS-EXT-AZ:availability_zone': u'nova',
        u'OS-EXT-SRV-ATTR:host': u'devstack',
        u'OS-EXT-SRV-ATTR:hypervisor_hostname': u'devstack',
        u'OS-EXT-SRV-ATTR:instance_name': u'instance-00000001',
        u'OS-EXT-STS:power_state': 1,
        u'OS-EXT-STS:task_state': None,
        u'OS-EXT-STS:vm_state': u'active',
        u'OS-SRV-USG:launched_at': created_now,
        u'OS-SRV-USG:terminated_at': None,
        u'accessIPv4': u'',
        u'accessIPv6': u'',
        u'addresses': {
            u'public': [{
                u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                u'OS-EXT-IPS:type': u'fixed',
                u'addr': u'172.25.0.3',
                u'version': 4
            }, {
                u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                u'OS-EXT-IPS:type': u'fixed',
                u'addr': u'2001:db8::3',
                u'version': 6
            }]
        },
        u'config_drive': u'True',
        u'created': created_now,
        u'flavor': {
            u'id': u'1',
            u'links': [{
                u'href': u'http://localhost:8774/dontcare',
                u'rel': u'bookmark'
            }]
        },
        u'hostId': u'd86d2c042a1f233227f70c5e9d2c5829de98d222d0922f469054ac17',
        u'host_name': u'devstack',
        u'id': instance_id,
        u'image': {
            u'id': u'46b77e67-ce40-44ca-823d-e6f83489f21e',
            u'links': [{
                u'href': u'http://localhost:8774/dontcare',
                u'rel': u'bookmark'
            }]
        },
        u'key_name': u'key',
        u'links': [
            {
                u'href': u'http://localhost:8774/dontcare',
                u'rel': u'self'
            },
            {
                u'href': u'http://localhost:8774/dontcare',
                u'rel': u'bookmark'
            }
        ],
        u'metadata': {},
        u'name': name,
        u'os-extended-volumes:volumes_attached': [],
        u'progress': 0,
        u'security_groups': [{u'name': u'default'}],
        u'status': u'active',
        u'tenant_id': tenant_id,
        u'updated': updated_now,
        u'user_id': USER1}

    attrs.update(kwargs)
    server = mock.Mock(spec=novaclient_servers.Server, **attrs)
    server.to_dict.return_value = attrs
    return server


class TestServerLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestServerLoaderPlugin, self).setUp()
        # Use unversioned notifications
        version_notifications = \
            'searchlight.elasticsearch.plugins.nova.notification_handler'\
            '.InstanceHandler._use_versioned_notifications'
        mock_versioned = mock.patch(version_notifications,
                                    return_value=False)
        mock_versioned.start()
        self.addCleanup(mock_versioned.stop)

        self.plugin = servers_plugin.ServerIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        self.instance1 = _instance_fixture(
            ID1, u'instance1', tenant_id=TENANT1,
            flavor=flavor1, image=imagea, addresses=net_ipv4,
            **{u'OS-EXT-AZ:availability_zone': u'az1',
               u'OS-EXT-SRV-ATTR:host': u'host1',
               u'hostId': u'host1'})
        self.instance2 = _instance_fixture(
            ID2, 'instance2', tenant_id=TENANT2,
            flavor=flavor2, image=imageb, addresses=net_ip4_6)
        self.instances = [self.instance1, self.instance2]

    def test_index_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Nova::Server', self.plugin.get_document_type())

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_serialize(self, mock_version):
        expected = {
            u'OS-DCF:diskConfig': u'MANUAL',
            u'OS-EXT-AZ:availability_zone': u'az1',
            u'OS-EXT-SRV-ATTR:host': u'host1',
            u'OS-EXT-SRV-ATTR:hypervisor_hostname': u'devstack',
            u'OS-EXT-SRV-ATTR:instance_name': u'instance-00000001',
            u'OS-EXT-STS:power_state': 1,
            u'OS-EXT-STS:task_state': None,
            u'OS-EXT-STS:vm_state': u'active',
            u'OS-SRV-USG:launched_at': created_now,
            u'OS-SRV-USG:terminated_at': None,
            u'accessIPv4': u'',
            u'accessIPv6': u'',
            u'addresses': {
                u'public': [{
                    u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                    u'OS-EXT-IPS:type': u'fixed',
                    u'addr': u'172.25.0.3',
                    u'version': 4
                }, {
                    u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                    u'OS-EXT-IPS:type': u'fixed',
                    u'addr': u'2001:db8::3',
                    u'version': 6
                }]
            },
            u'config_drive': u'True',
            u'flavor': {u'id': u'1'},
            u'hostId': u'host1',
            u'host_name': u'devstack',
            u'id': u'6c41b4d1-f0fa-42d6-9d8d-e3b99695aa69',
            u'image': {u'id': u'a'},
            u'key_name': u'key',
            u'metadata': {},
            u'name': u'instance1',
            u'os-extended-volumes:volumes_attached': [],
            u'owner': u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0',
            u'security_groups': [u'default'],
            u'status': u'active',
            u'tenant_id': u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0',
            u'project_id': u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0',
            u'updated': updated_now,
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55',
            u'networks': [{
                u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                u'version': 4,
                u'ipv4_addr': u'127.0.0.1',
                u'OS-EXT-IPS:type': u'fixed',
                u'name': u'net4',
            }],
            u'addresses': net_ipv4,
            u'created': created_now,
            u'created_at': created_now,
            u'updated': updated_now,
            u'updated_at': updated_now,
        }
        with mock.patch(nova_server_getter, return_value=self.instance1):
            serialized = self.plugin.serialize(self.instance1.id)

        self.assertEqual(expected, serialized)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_serialize_no_image(self, mock_version):
        instance = _instance_fixture(
            ID3, u'instance3', tenant_id=TENANT1,
            flavor=flavor1, image='', addresses=net_ipv4,
            **{u'OS-EXT-AZ:availability_zone': u'az1',
               u'OS-EXT-SRV-ATTR:host': u'host1',
               u'hostId': u'host1'})
        expected = {
            u'OS-DCF:diskConfig': u'MANUAL',
            u'OS-EXT-AZ:availability_zone': u'az1',
            u'OS-EXT-SRV-ATTR:host': u'host1',
            u'OS-EXT-SRV-ATTR:hypervisor_hostname': u'devstack',
            u'OS-EXT-SRV-ATTR:instance_name': u'instance-00000001',
            u'OS-EXT-STS:power_state': 1,
            u'OS-EXT-STS:task_state': None,
            u'OS-EXT-STS:vm_state': u'active',
            u'OS-SRV-USG:launched_at': created_now,
            u'OS-SRV-USG:terminated_at': None,
            u'accessIPv4': u'',
            u'accessIPv6': u'',
            u'addresses': {
                u'public': [{
                    u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                    u'OS-EXT-IPS:type': u'fixed',
                    u'addr': u'172.25.0.3',
                    u'version': 4
                }, {
                    u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                    u'OS-EXT-IPS:type': u'fixed',
                    u'addr': u'2001:db8::3',
                    u'version': 6
                }]
            },
            u'config_drive': u'True',
            u'flavor': {u'id': u'1'},
            u'hostId': u'host1',
            u'host_name': u'devstack',
            u'id': u'a380287d-1f61-4887-959c-8c5ab8f75f8f',
            u'key_name': u'key',
            u'metadata': {},
            u'name': u'instance3',
            u'os-extended-volumes:volumes_attached': [],
            u'owner': u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0',
            u'project_id': u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0',
            u'security_groups': [u'default'],
            u'status': u'active',
            u'tenant_id': u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0',
            u'updated': updated_now,
            u'user_id': u'27f4d76b-be62-4e4e-aa33bb11cc55',
            u'networks': [{
                u'OS-EXT-IPS-MAC:mac_addr': u'fa:16:3e:1e:37:32',
                u'version': 4,
                u'ipv4_addr': u'127.0.0.1',
                u'OS-EXT-IPS:type': u'fixed',
                u'name': u'net4',
            }],
            u'addresses': net_ipv4,
            u'created': created_now,
            u'created_at': created_now,
            u'updated': updated_now,
            u'updated_at': updated_now,
        }
        with mock.patch(nova_server_getter, return_value=instance):
            serialized = self.plugin.serialize(instance.id)

        self.assertEqual(expected, serialized)

    def test_facets_non_admin(self):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine
        # Don't care about the actual aggregation result; the base functions
        # are tested separately.
        mock_engine.search.return_value = {'aggregations': {},
                                           'hits': {'total': 0}}

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=False
        )

        facets, _ = self.plugin.get_facets(fake_request.context)
        facet_names = [f['name'] for f in facets]

        network_facets = ('name', 'version', 'ipv6_addr', 'ipv4_addr',
                          'OS-EXT-IPS-MAC:mac_addr', 'OS-EXT-IPS:type')
        expected_facet_names = [
            'OS-EXT-AZ:availability_zone',
            'created_at', 'description',
            'flavor.id', 'id', 'image.id', 'locked', 'name',
            'owner', 'security_groups', 'status', 'tags', 'updated_at',
            'user_id', 'OS-EXT-STS:vm_state']
        expected_facet_names.extend(['networks.' + f for f in network_facets])

        self.assertEqual(set(expected_facet_names), set(facet_names))

        # Test fields with options
        complex_facet_option_fields = (
            'networks.name', 'networks.OS-EXT-IPS:type', 'networks.version')
        aggs = dict(unit_test_utils.complex_facet_field_agg(name)
                    for name in complex_facet_option_fields)

        simple_facet_option_fields = (
            'status', 'OS-EXT-AZ:availability_zone', 'security_groups',
            'locked', 'flavor.id', 'image.id'
        )
        aggs.update(dict(unit_test_utils.simple_facet_field_agg(name)
                         for name in simple_facet_option_fields))

        expected_agg_query = {
            'aggs': aggs,
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {ROLE_USER_FIELD: 'user'}},
                            'should': [{'term': {'tenant_id': TENANT1}}],
                            'minimum_should_match': 1
                        }
                    }
                }
            }
        }
        mock_engine.search.assert_called_with(
            index=self.plugin.alias_name_search,
            doc_type=self.plugin.get_document_type(),
            body=expected_agg_query,
            ignore_unavailable=True,
            size=0
        )

    def test_facets_admin(self):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine
        mock_engine.search.return_value = {'aggregations': {},
                                           'hits': {'total': 0}}

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=True
        )

        facets, _ = self.plugin.get_facets(fake_request.context)
        facet_names = [f['name'] for f in facets]

        network_facets = ('name', 'version', 'ipv6_addr', 'ipv4_addr',
                          'OS-EXT-IPS-MAC:mac_addr', 'OS-EXT-IPS:type')
        expected_facet_names = [
            'OS-EXT-SRV-ATTR:hypervisor_hostname',
            'OS-EXT-AZ:availability_zone', 'created_at', 'description',
            'flavor.id', 'host_status', 'id', 'image.id', 'locked',
            'name', 'owner', 'project_id', 'security_groups', 'status',
            'tags', 'tenant_id', 'updated_at', 'user_id',
            'OS-EXT-STS:vm_state']
        expected_facet_names.extend(['networks.' + f for f in network_facets])

        self.assertEqual(set(expected_facet_names), set(facet_names))

        complex_facet_option_fields = (
            'networks.name', 'networks.OS-EXT-IPS:type', 'networks.version')
        aggs = dict(unit_test_utils.complex_facet_field_agg(name)
                    for name in complex_facet_option_fields)

        simple_facet_option_fields = (
            'status', 'OS-EXT-AZ:availability_zone', 'security_groups',
            'host_status', 'locked', 'image.id', 'flavor.id'
        )
        aggs.update(dict(unit_test_utils.simple_facet_field_agg(name)
                         for name in simple_facet_option_fields))

        expected_agg_query = {
            'aggs': aggs,
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {ROLE_USER_FIELD: 'admin'}},
                            'should': [{'term': {'tenant_id': TENANT1}}],
                            'minimum_should_match': 1
                        }
                    }
                }
            }
        }
        mock_engine.search.assert_called_with(
            index=self.plugin.alias_name_search,
            doc_type=self.plugin.get_document_type(),
            body=expected_agg_query,
            ignore_unavailable=True,
            size=0
        )

    def test_facets_all_projects(self):
        # For non admins, all_projects should have no effect
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        # Don't really care about the return values
        mock_engine.search.return_value = {
            'aggregations': {},
            'hits': {'total': 0}
        }

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=False
        )

        self.plugin.get_facets(fake_request.context, all_projects=True)

        expected_agg_query_filter = {
            'bool': {
                'filter': {
                    'bool': {
                        'must': {'term': {ROLE_USER_FIELD: 'user'}},
                        'should': [{'term': {'tenant_id': TENANT1}}],
                        'minimum_should_match': 1
                    }
                }
            }
        }

        # Call args are a tuple (name, posargs, kwargs)
        search_call = mock_engine.search.mock_calls[0]
        search_call_body = search_call[2]['body']
        self.assertEqual(set(['aggs', 'query']), set(search_call_body.keys()))

        # The aggregation query's tested elsewhere, just check the query filter
        self.assertEqual(expected_agg_query_filter, search_call_body['query'])

        # Admins can request all_projects
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=True
        )

        self.plugin.get_facets(fake_request.context, all_projects=True)

        # Test the SECOND call to the mock
        search_call = mock_engine.search.mock_calls[1]
        search_call_body = search_call[2]['body']

        # We don't expect any filter query here, just the aggregations.
        self.assertEqual(set(['aggs', 'query']), set(search_call_body.keys()))

    def test_facets_no_mapping(self):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=True
        )

        mock_engine.search.return_value = {
            'aggregations': {
                'status': {'buckets': []},
                'image.id': {'doc_count': 0}
            },
            'hits': {'total': 0}
        }

        facets, _ = self.plugin.get_facets(fake_request.context)

        status_facet = list(filter(lambda f: f['name'] == 'status',
                                   facets))[0]
        image_facet = list(filter(lambda f: f['name'] == 'image.id',
                                  facets))[0]
        expected_status = {'name': 'status', 'options': [], 'type': 'string'}
        expected_image = {'name': 'image.id', 'options': [], 'type': 'string',
                          'resource_type': 'OS::Glance::Image',
                          'nested': False}

        self.assertEqual(expected_status, status_facet)
        self.assertEqual(expected_image, image_facet)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_created_at_updated_at(self, mock_version):
        self.assertNotIn('created_at', self.instance1.to_dict())
        self.assertNotIn('updated_at', self.instance1.to_dict())

        with mock.patch(nova_server_getter, return_value=self.instance1):
            serialized = self.plugin.serialize(self.instance1.id)

        self.assertEqual(serialized['created_at'], created_now)
        self.assertEqual(serialized['updated_at'], updated_now)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_update_404_deletes(self, mock_version):
        """Test that if a server is missing on a notification event, it
        gets deleted from the index
        """
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        notification_handler = self.plugin.get_notification_handler()
        doc_deleter = self.plugin._index_helper
        nova_exc = novaclient.exceptions.NotFound('missing')
        with mock.patch(nova_server_getter,
                        side_effect=nova_exc) as mock_get:
            with mock.patch.object(doc_deleter,
                                   'delete_document') as mock_deleter:
                fake_timestamp = '2015-09-01 08:57:35.282586'
                notification_handler.index_from_api(
                    'fake_event_type',
                    {u'instance_id': u'missing',
                     u'updated_at': u'2016-02-01T00:00:00Z'},
                    fake_timestamp)
                mock_get.assert_called_once_with(u'missing')
                mock_deleter.assert_called_once_with(
                    {'_id': u'missing', '_version': 454284800097855282})

    def test_vol_events_supported(self):
        not_handler = self.plugin.get_notification_handler()
        vol_payload = {
            "instance_id": "a",
            "volume_id": "b"
        }

        with mock.patch.object(not_handler, 'index_from_api') as mock_update:
            events = not_handler.get_event_handlers()
            self.assertIn('compute.instance.volume.attach', events)
            self.assertIn('compute.instance.volume.detach', events)

            attach = events['compute.instance.volume.attach']
            attach(vol_payload, 1234)

            mock_update.assert_called_with(vol_payload, 1234)
            mock_update.reset()

            detach = events['compute.instance.volume.detach']
            detach(vol_payload, 1234)

            mock_update.assert_called_with(vol_payload, 1234)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_filter_result(self, mock_version):
        """We reformat outgoing results so that security group looks like the
        response we get from the nova API.
        """
        with mock.patch(nova_server_getter, return_value=self.instance1):
            es_result = self.plugin.serialize(self.instance1.id)
        self.plugin.filter_result({'_source': es_result}, None)
        self.assertEqual([{'name': 'default'}], es_result['security_groups'])

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_boot_state_change_notifications(self, mock_version):
        """This is an attempt to stop hammering the nova API!
        Many state change events are received in a short space of time during
        nova instance boots. We'll ignore some of them.
        """
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        successful_boot_events = [
            ("compute.instance.update",
             dict(state_description="scheduling", old_state="building",
                  state="building", old_task_state="scheduling",
                  new_task_state="scheduling"),
             "2016-05-12 20:44:40.352048"),
            ("compute.instance.update",
             dict(state_description="scheduling", old_state="building",
                  state="building", old_task_state="scheduling",
                  new_task_state=None),
             "2016-05-12 20:44:40.867902"),
            ("compute.instance.create.start",
             dict(state_description="", state="building"),
             "2016-05-12 20:44:40.885366"),
            ("compute.instance.update",
             dict(state_description="", old_state="building",
                  state="building", old_task_state=None,
                  new_task_state=None),
             "2016-05-12 20:44:40.981856"),
            ("compute.instance.update",
             dict(state_description="networking", old_state="building",
                  state="building", old_task_state="networking",
                  new_task_state="networking"),
             "2016-05-12 20:44:41.082192"),
            ("compute.instance.update",
             dict(state_description="block_device_mapping",
                  old_state="building", state="building",
                  old_task_state="networking",
                  new_task_state="block_device_mapping"),
             "2016-05-12 20:44:41.298666"),
            ("compute.instance.update",
             dict(state_description="spawning", old_state="building",
                  state="building", old_task_state="block_device_mapping",
                  new_task_state="spawning"),
             "2016-05-12 20:44:41.355864"),
            ("port.create.start",
             dict(device_owner="compute:None",
                  device_id="913a1d19-f96f-4bb2-b459-84dd25ebb923"),
             "2016-05-12 20:44:41.673135"),
            ("port.create.end",
             dict(device_owner="compute:None",
                  device_id="913a1d19-f96f-4bb2-b459-84dd25ebb923"),
             "2016-05-12 20:44:42.700366"),
            ("image.send", {}, "2016-05-12 20:44:42.700366"),
            ("image.send", {}, "2016-05-12 20:44:45.444854"),
            ("compute.instance.update",
             dict(state_description="", old_state="building",
                  state="active", old_task_state="spawning",
                  new_task_state=None),
             "2016-05-12 20:44:52.915259"),
            ("compute.instance.create.end",
             dict(state_description="", state="active"),
             "2016-05-12 20:44:52.976685"),
        ]

        for event in successful_boot_events:
            event[1]["instance_id"] = u"a380287d-1f61-4887-959c-8c5ab8f75f8f"

        # Feed the events to the notification handler
        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        def handle_message(message, expected_event_type):
            self.assertEqual(expected_event_type, message[0],
                             "Expected event type doesn't match test "
                             "message type.")
            # type, payload, timestamp
            type_handler = event_handlers.get(message[0], None)
            if type_handler:
                type_handler(*message)

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                def assert_call_counts(get, save):
                    # Helper to reduce copy pasta
                    self.assertEqual(get, nova_getter.call_count)
                    self.assertEqual(save, mock_save.call_count)

                # Expect a nova API hit from the first update
                handle_message(successful_boot_events[0],
                               "compute.instance.update")
                assert_call_counts(get=1, save=1)

                # Expect nothing for the next update (should be ignored)
                handle_message(successful_boot_events[1],
                               "compute.instance.update")
                assert_call_counts(get=1, save=1)

                # Causes a nova GET
                handle_message(successful_boot_events[2],
                               "compute.instance.create.start")
                assert_call_counts(get=2, save=2)

                # Update after create.start is ignored
                handle_message(successful_boot_events[3],
                               "compute.instance.update")
                assert_call_counts(get=2, save=2)

                # Then networking, disk mapping etc
                mock_engine.get.return_value = {
                    '_version': 3,
                    '_source': {
                        'OS-EXT-STS:vm_state': 'building',
                        'OS-EXT-STS:task_state': None,
                        'id': 1
                    }
                }
                handle_message(successful_boot_events[4],
                               "compute.instance.update")
                assert_call_counts(get=2, save=3)

                mock_engine.get.return_value = {
                    '_version': 4,
                    '_source': {
                        'OS-EXT-STS:vm_state': 'building',
                        'OS-EXT-STS:task_state': 'networking',
                        'id': 1
                    }
                }
                handle_message(successful_boot_events[5],
                               "compute.instance.update")
                assert_call_counts(get=2, save=4)

                mock_engine.get.return_value = {
                    '_version': 5,
                    '_source': {
                        'OS-EXT-STS:vm_state': 'building',
                        'OS-EXT-STS:task_state': 'block_device_mapping',
                        'id': 1
                    }
                }
                handle_message(successful_boot_events[6],
                               "compute.instance.update")
                assert_call_counts(get=2, save=5)

                # port.create events are ignored
                handle_message(successful_boot_events[7],
                               "port.create.start")
                handle_message(successful_boot_events[8],
                               "port.create.end")
                assert_call_counts(get=2, save=5)

                # image.send ignored
                handle_message(successful_boot_events[9], "image.send")
                handle_message(successful_boot_events[10], "image.send")
                assert_call_counts(get=2, save=5)

                # final update is also ignored because create.end follows
                handle_message(successful_boot_events[11],
                               "compute.instance.update")
                assert_call_counts(get=2, save=5)

                # create.end causes full save
                handle_message(successful_boot_events[12],
                               "compute.instance.create.end")
                assert_call_counts(get=3, save=6)

    def test_racing_state_change_notifications(self):
        """Test that a 'state update' change doesn't get applied if it looks
        like the instance state has already been updated by a later change.
        """
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        with mock.patch.object(self.plugin.index_helper,
                               'save_document') as mock_save:
            # Simulate the 'spawning' event getting processed first
            mock_engine.get.return_value = {
                '_version': 4,
                '_source': {
                    'OS-EXT-STS:vm_state': 'building',
                    'OS-EXT-STS:task_state': 'spawning'
                }
            }
            event_handlers['compute.instance.update'](
                'compute.instance.update',
                dict(state_description="block_device_mapping",
                     old_state="building", state="building",
                     old_task_state="networking",
                     new_task_state="block_device_mapping",
                     instance_id=u"a380287d-1f61-4887-959c-8c5ab8f75f8f"),
                "2016-05-12 20:44:41.298666")

            self.assertEqual(0, mock_save.call_count)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_delete_state_change_notifications(self, mock_version):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        delete_events = [
            ('compute.instance.update',
             dict(state_description='deleting', state='active',
                  old_task_state='deleting', new_task_state='deleting'),
             '2016-05-23 18:44:02.889647'),
            ('compute.instance.delete.start',
             dict(state_description='deleting', state='active'),
             '2016-05-23 18:44:02.943141'),
            ('compute.instance.shutdown.start',
             dict(state_description='deleting', state='active'),
             '2016-05-23 18:44:02.950168'),
            ('compute.instance.update',
             dict(state_description='deleting', state='active',
                  old_state='active', new_task_state='deleting',
                  old_task_state='deleting'),
             '2016-05-23 18:44:04.155735'),
            ('port.delete.start', {}, '2016-05-23 18:44:04.435210'),
            ('port.delete.end', {}, '2016-05-23 18:44:04.627044'),
            ('compute.instance.shutdown.end',
             dict(state_description='deleting', state='active'),
             '2016-05-23 18:44:04.696012'),
            ('compute.instance.update',
             dict(state_description='', state='deleted', new_task_state=None,
                  old_task_state='deleting'),
             '2016-05-23 18:44:04.779076'),
            ('compute.instance.delete.end',
             dict(state_description='', state='deleted',
                  updated_at='2016-05-23 18:44:04'),
             '2016-05-23 18:44:04.916681')
        ]

        for event in delete_events:
            event[1]["instance_id"] = u"a380287d-1f61-4887-959c-8c5ab8f75f8f"

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        def handle_message(message, expected_event_type, expect_handled=True):
            self.assertEqual(expected_event_type, message[0],
                             "Expected event type doesn't match test "
                             "message type.")
            # type, payload, timestamp
            type_handler = event_handlers.get(message[0], None)
            if type_handler:
                type_handler(*message)
            else:
                if expect_handled:
                    self.fail("Expected event '%s' to be handled" %
                              expected_event_type)

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch.object(self.plugin.index_helper,
                              'delete_document') as mock_delete, \
            mock.patch.object(self.plugin.index_helper,
                              'get_document') as mock_es_getter, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                def assert_call_counts(get=0, save=0, es_gets=0):
                    # Helper to reduce copy pasta
                    self.assertEqual(get, nova_getter.call_count)
                    self.assertEqual(save, mock_save.call_count)
                    self.assertEqual(es_gets, mock_es_getter.call_count)

                handle_message(delete_events[0],
                               "compute.instance.update")
                assert_call_counts(es_gets=1)

                handle_message(delete_events[1],
                               "compute.instance.delete.start",
                               expect_handled=False)

                handle_message(delete_events[2],
                               "compute.instance.shutdown.start",
                               expect_handled=False)
                assert_call_counts(es_gets=1)

                handle_message(delete_events[3],
                               "compute.instance.update")
                assert_call_counts(es_gets=2)

                # Ignore port events
                handle_message(delete_events[4],
                               "port.delete.start",
                               expect_handled=False)
                handle_message(delete_events[5],
                               "port.delete.end",
                               expect_handled=False)

                handle_message(delete_events[6],
                               "compute.instance.shutdown.end")
                assert_call_counts(get=1, save=1, es_gets=2)

                # Ignore this one too
                handle_message(delete_events[7],
                               "compute.instance.update")
                assert_call_counts(get=1, save=1, es_gets=2)

                handle_message(delete_events[8],
                               "compute.instance.delete.end")
                assert_call_counts(get=1, save=1, es_gets=2)
                self.assertEqual(1, mock_delete.call_count)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_pause_state_change_notifications(self, mock_version):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        pause_events = [
            ('compute.instance.update',
             dict(state_description='pausing', state='active',
                  old_task_state='pausing', new_task_state='pausing'),
             '2016-06-17 19:52:13.523135'),
            ('compute.instance.pause.start',
             dict(state_description='pausing', state='active'),
             '2016-06-17 19:52:13.628180'),
            ('compute.instance.update',
             dict(state_description='', state='paused',
                  old_task_state='pausing', new_task_state=None),
             '2016-06-17 19:52:13.771604'),
            ('compute.instance.pause.end',
             dict(state_description='', state='paused'),
             '2016-06-17 19:52:13.808362')
        ]

        for event in pause_events:
            event[1]["instance_id"] = u"a380287d-1f61-4887-959c-8c5ab8f75f8f"

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        def handle_message(message, expected_event_type, expect_handled=True):
            self.assertEqual(expected_event_type, message[0],
                             "Expected event type doesn't match test "
                             "message type.")
            # type, payload, timestamp
            type_handler = event_handlers.get(message[0], None)
            if type_handler:
                type_handler(*message)
            else:
                if expect_handled:
                    self.fail("Expected event '%s' to be handled" %
                              expected_event_type)

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch.object(self.plugin.index_helper,
                              'get_document') as mock_es_getter, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                def assert_call_counts(get=0, save=0, es_gets=0):
                    # Helper to reduce copy pasta
                    self.assertEqual(get, nova_getter.call_count)
                    self.assertEqual(save, mock_save.call_count)
                    self.assertEqual(es_gets, mock_es_getter.call_count)

                handle_message(pause_events[0],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(pause_events[1],
                               "compute.instance.pause.start",
                               expect_handled=False)
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(pause_events[2],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(pause_events[3],
                               "compute.instance.pause.end",
                               expect_handled=False)
                assert_call_counts(get=1, save=1, es_gets=1)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_reboot_state_change_notifications(self, mock_version):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        reboot_events = [
            ('compute.instance.update',
             dict(state_description='rebooting', state='active',
                  old_task_state='rebooting', new_task_state='rebooting'),
             '2016-07-17 19:52:13.523135'),
            ('compute.instance.reboot.start',
             dict(state_description='rebooting', state='active'),
             '2016-07-17 19:52:13.628180'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='rebooting', new_task_state='reboot_pending'),
             '2016-07-17 19:52:13.771604'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='reboot_pending',
                  new_task_state='reboot_started'),
             '2016-07-17 19:52:13.781605'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='reboot_started', new_task_state=None),
             '2016-07-17 19:52:13.791605'),
            ('compute.instance.reboot.end',
             dict(state_description='', state='active'),
             '2016-07-17 19:52:13.808362')
        ]

        for event in reboot_events:
            event[1]["instance_id"] = u"a380287d-1f61-4887-959c-8c5ab8f75f8f"

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        def handle_message(message, expected_event_type, expect_handled=True):
            self.assertEqual(expected_event_type, message[0],
                             "Expected event type doesn't match test "
                             "message type.")
            # type, payload, timestamp
            type_handler = event_handlers.get(message[0], None)
            if type_handler:
                type_handler(*message)
            else:
                if expect_handled:
                    self.fail("Expected event '%s' to be handled" %
                              expected_event_type)

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch.object(self.plugin.index_helper,
                              'get_document') as mock_es_getter, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                def assert_call_counts(get=0, save=0, es_gets=0):
                    # Helper to reduce copy pasta
                    self.assertEqual(get, nova_getter.call_count)
                    self.assertEqual(save, mock_save.call_count)
                    self.assertEqual(es_gets, mock_es_getter.call_count)

                handle_message(reboot_events[0],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(reboot_events[1],
                               "compute.instance.reboot.start",
                               expect_handled=False)
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(reboot_events[2],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=2)

                handle_message(reboot_events[3],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=3)

                handle_message(reboot_events[4],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=3)

                handle_message(reboot_events[5],
                               "compute.instance.reboot.end",
                               expect_handled=False)
                assert_call_counts(get=1, save=1, es_gets=3)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_shelve_state_change_notifications(self, mock_version):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        shelve_events = [
            ('compute.instance.update',
             dict(state_description='shelving', state='active',
                  old_task_state='shelving', new_task_state='shelving'),
             '2016-07-17 19:52:13.523135'),
            ('compute.instance.shelve.start',
             dict(state_description='shelving', state='active'),
             '2016-07-17 19:52:13.628180'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='shelving',
                  new_task_state='shelving_image_pending_upload'),
             '2016-07-17 19:52:13.771604'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='shelving_image_pending_upload',
                  new_task_state='shelving_image_uploading'),
             '2016-07-17 19:52:13.781605'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='shelving_image_uploading',
                  new_task_state='shelving_image_pending_upload'),
             '2016-07-17 19:52:13.791605'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='shelving_image_pending_upload',
                  new_task_state='shelving_image_uploading'),
             '2016-07-17 19:52:13.801605'),
            ('compute.instance.update',
             dict(state_description='', state='shelved',
                  old_task_state='shelving_image_uploading',
                  new_task_state='shelving_offloading'),
             '2016-07-17 19:52:13.811605'),
            ('compute.instance.update',
             dict(state_description='', state='shelved',
                  old_task_state='shelving_offloading',
                  new_task_state='shelving_offloading'),
             '2016-07-17 19:52:13.821605'),
            ('compute.instance.update',
             dict(state_description='', state='shelved_offloaded',
                  old_task_state='shelving_offloading',
                  new_task_state=None),
             '2016-07-17 19:52:13.831605'),
            ('compute.instance.shelve.end',
             dict(state_description='', state='active'),
             '2016-07-17 19:52:13.858362')
        ]

        for event in shelve_events:
            event[1]["instance_id"] = u"a380287d-1f61-4887-959c-8c5ab8f75f8f"

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        def handle_message(message, expected_event_type, expect_handled=True):
            self.assertEqual(expected_event_type, message[0],
                             "Expected event type doesn't match test "
                             "message type.")
            # type, payload, timestamp
            type_handler = event_handlers.get(message[0], None)
            if type_handler:
                type_handler(*message)
            else:
                if expect_handled:
                    self.fail("Expected event '%s' to be handled" %
                              expected_event_type)

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch.object(self.plugin.index_helper,
                              'get_document') as mock_es_getter, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                def assert_call_counts(get=0, save=0, es_gets=0):
                    # Helper to reduce copy pasta
                    self.assertEqual(get, nova_getter.call_count)
                    self.assertEqual(save, mock_save.call_count)
                    self.assertEqual(es_gets, mock_es_getter.call_count)

                handle_message(shelve_events[0],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(shelve_events[1],
                               "compute.instance.shelve.start",
                               expect_handled=False)
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(shelve_events[2],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=2)

                handle_message(shelve_events[3],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=3)

                handle_message(shelve_events[4],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=4)

                handle_message(shelve_events[5],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=5)

                handle_message(shelve_events[6],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=6)

                handle_message(shelve_events[7],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=6)

                handle_message(shelve_events[8],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=6)

                handle_message(shelve_events[9],
                               "compute.instance.shelve.end",
                               expect_handled=False)
                assert_call_counts(get=1, save=1, es_gets=6)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_unshelve_state_change_notifications(self, mock_version):
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        unshelve_events = [
            ('compute.instance.update',
             dict(state_description='unshelving', state='shelved_offloaded',
                  old_task_state='unshelving', new_task_state='unshelving'),
             '2016-07-17 19:52:13.523135'),
            ('compute.instance.unshelve.start',
             dict(state_description='unshelving', state='shelved_offloaded'),
             '2016-07-17 19:52:13.628180'),
            ('compute.instance.update',
             dict(state_description='', state='shelved_offloaded',
                  old_task_state='unshelving', new_task_state='unshelving'),
             '2016-07-17 19:52:13.771604'),
            ('compute.instance.update',
             dict(state_description='', state='shelved_offloaded',
                  old_task_state='unshelving',
                  new_task_state='spawning'),
             '2016-07-17 19:52:13.781605'),
            ('compute.instance.update',
             dict(state_description='', state='shelved_offloaded',
                  old_task_state='spawning', new_task_state='spawning'),
             '2016-07-17 19:52:13.791605'),
            ('compute.instance.update',
             dict(state_description='', state='active',
                  old_task_state='spawning', new_task_state=None),
             '2016-07-17 19:52:13.801605'),
            ('compute.instance.unshelve.end',
             dict(state_description='', state='active'),
             '2016-07-17 19:52:13.818362')
        ]

        for event in unshelve_events:
            event[1]["instance_id"] = u"a380287d-1f61-4887-959c-8c5ab8f75f8f"

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        def handle_message(message, expected_event_type, expect_handled=True):
            self.assertEqual(expected_event_type, message[0],
                             "Expected event type doesn't match test "
                             "message type.")
            # type, payload, timestamp
            type_handler = event_handlers.get(message[0])
            if type_handler:
                type_handler(*message)
            else:
                if expect_handled:
                    self.fail("Expected event '%s' to be handled" %
                              expected_event_type)

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch.object(self.plugin.index_helper,
                              'get_document') as mock_es_getter, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                def assert_call_counts(get=0, save=0, es_gets=0):
                    # Helper to reduce copy pasta
                    self.assertEqual(get, nova_getter.call_count)
                    self.assertEqual(save, mock_save.call_count)
                    self.assertEqual(es_gets, mock_es_getter.call_count)

                handle_message(unshelve_events[0],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(unshelve_events[1],
                               "compute.instance.unshelve.start",
                               expect_handled=False)
                assert_call_counts(get=0, save=0, es_gets=1)

                handle_message(unshelve_events[2],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=2)

                handle_message(unshelve_events[3],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=3)

                handle_message(unshelve_events[4],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=3)

                handle_message(unshelve_events[5],
                               "compute.instance.update")
                assert_call_counts(get=0, save=0, es_gets=3)

                handle_message(unshelve_events[6],
                               "compute.instance.unshelve.end",
                               expect_handled=False)
                assert_call_counts(get=1, save=1, es_gets=3)

    @mock.patch(nova_version_getter, return_value=fake_version_list)
    def test_server_rename(self, mock_version):
        """Test that a notification received with an 'active' state
        is processed.
        """
        instance_id = self.instance1.id
        update_event = \
            dict(old_state='active', state='active',
                 old_task_state=None, new_task_state=None,
                 name='renamed', instance_id=instance_id)

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()
        type_handler = event_handlers['compute.instance.update']

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save, \
            mock.patch(nova_server_getter,
                       return_value=self.instance1) as nova_getter:

                type_handler('compute.instance.update', update_event,
                             '2016-03-17 19:52:13.523135')
                nova_getter.assert_called_with(instance_id)
                self.assertEqual(1, mock_save.call_count)


class TestVersionedServerNotifications(test_utils.BaseTestCase):
    def setUp(self):
        super(TestVersionedServerNotifications, self).setUp()

        self.plugin = servers_plugin.ServerIndex()

    def test_versioned_create(self):
        create_event = {
            "nova_object.name": "InstanceActionPayload",
            "nova_object.namespace": "nova",
            "nova_object.version": "1.1",
            "nova_object.data": {
                "image_uuid": "155d900f-4e14-4e4c-a73d-069cbf4541e6",
                "tenant_id": "6f70656e737461636b20342065766572",
                "created_at": "2017-03-17T19:52:13Z",
                "display_name": "some-server",
                "display_description": "some-server",
                "state": "active",
                "flavor": {
                    "nova_object.name": "FlavorPayload",
                    "nova_object.data": {
                        "flavorid": "a22d5517-147c-4147-a0d1-e698df5cd4e3",
                        "name": "test_flavor",
                        "root_gb": 1,
                        "vcpus": 1,
                        "ephemeral_gb": 0,
                        "memory_mb": 512
                    }
                },
                "uuid": "178b0921-8f85-4257-88b6-2e743b5a975c",
                "power_state": "running",
                "ip_addresses": [{
                    "nova_object.name": "IpPayload",
                    "nova_object.data": {
                        "mac": "fa:16:3e:4c:2c:30",
                        "address": "192.168.1.3",
                        "port_uuid": "ce531f90-199f-48c0-816c-13e38010b442",
                        "version": 4,
                        "label": "private-network",
                        "device_name": "tapce531f90-19"
                    }
                }],
                'OS-EXT-STS:vm_state': 'active',
            }
        }

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        expected = {
            'image': {'id': "155d900f-4e14-4e4c-a73d-069cbf4541e6"},
            'tenant_id': "6f70656e737461636b20342065766572",
            'project_id': "6f70656e737461636b20342065766572",
            "created_at": "2017-03-17T19:52:13Z",
            "name": "some-server",
            "description": "some-server",
            "flavor": {"id": "a22d5517-147c-4147-a0d1-e698df5cd4e3"},
            "id": "178b0921-8f85-4257-88b6-2e743b5a975c",
            "power_state": "running",
            "status": "active",
            'OS-EXT-STS:vm_state': 'active',
            "networks": [{
                "version": 4,
                "ipv4_addr": "192.168.1.3",
                "OS-EXT-IPS-MAC:mac_addr": "fa:16:3e:4c:2c:30",
                "name": "tapce531f90-19"
            }],
        }
        expected_version = 489780333780333818

        with mock.patch.object(self.plugin.index_helper,
                               'save_documents') as mock_save:
            handler = event_handlers.get('instance.create.end')
            self.assertIsNotNone(handler)
            handler(payload=create_event,
                    event_type='instance.create.end',
                    timestamp='2017-03-17 19:52:13.818362')

            self.assertEqual(1, mock_save.call_count)
            mock_save.assert_called_with([expected], [expected_version])



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from searchlight.common import utils
from searchlight.elasticsearch.plugins.neutron import\
    routers as routers_plugin
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


_now_str = utils.isotime(datetime.datetime.utcnow())
USER1 = u'27f4d76b-be62-4e4e-aa33bb11cc55'
ID1 = "813dd936-663e-4e5b-877c-986021b73e2c"
NETID1 = "98dcb60c-59b9-4b3c-bf6a-b8504112e978"
TENANT1 = "8eaac046b2c44ab99246cb0850c7f06d"


def _router_fixture(router_id, tenant_id, name, **kwargs):
    fixture = {
        u'admin_state_up': True,
        u'availability_zone_hints': [],
        u'availability_zones': [],
        u'external_gateway_info': {
            u'enable_snat': True,
            u'external_fixed_ips': [{
                u'ip_address': u'172.25.0.2',
                u'subnet_id': u'8cbe9b71-ecf1-4355-b5ba-dee54ec88fa7'
            }, {
                u'ip_address': u'2001:db8::1',
                u'subnet_id': u'd7774648-ba81-477a-9329-2acc9a810e50'
            }],
            u'network_id': u'8891323e-bf5b-48d7-a75e-669af0608538'},
        u'id': router_id,
        u'name': name,
        u'routes': [],
        u'status': u'ACTIVE',
        u'tenant_id': tenant_id}
    fixture.update(**kwargs)
    return fixture


def _gateway_info(network_id, fixed_ips):
    return {
        u'enable_snat': True,
        u'external_fixed_ips': fixed_ips,
        u'network_id': network_id
    }


class TestRouterLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestRouterLoaderPlugin, self).setUp()
        self.plugin = routers_plugin.RouterIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        self.router1 = _router_fixture(router_id=ID1, network_id=NETID1,
                                       tenant_id=TENANT1, name="test-router-1")
        self.routers = [self.router1]

    def test_admin_only(self):
        self.assertEqual(['distributed', 'ha'], self.plugin.admin_only_fields)

    def test_document_type(self):
        self.assertEqual('OS::Neutron::Router',
                         self.plugin.get_document_type())

    def test_rbac_filter(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=False
        )
        rbac_terms = self.plugin._get_rbac_field_filters(fake_request.context)
        self.assertEqual(
            [{"term": {"tenant_id": TENANT1}}],
            rbac_terms
        )

    def test_notification_events(self):
        handler = self.plugin.get_notification_handler()
        self.assertEqual(
            set(['router.create.end', 'router.update.end',
                 'router.delete.end']),
            set(handler.get_event_handlers().keys())
        )



# Copyright 2011, OpenStack Foundation
# Copyright 2012, Red Hat, Inc.
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_config import cfg
import oslo_messaging
import six


notifier_opts = [
    cfg.StrOpt('default_publisher_id', default="image.localhost",
               help='Default publisher_id for outgoing notifications.'),
    cfg.ListOpt('disabled_notifications', default=[],
                help='List of disabled notifications. A notification can be '
                     'given either as a notification type to disable a single '
                     'event, or as a notification group prefix to disable all '
                     'events within a group. Example: if this config option '
                     'is set to ["image.create", "metadef_namespace"], then '
                     '"image.create" notification will not be sent after '
                     'image is created and none of the notifications for '
                     'metadefinition namespaces will be sent.'),
]

CONF = cfg.CONF
CONF.register_opts(notifier_opts)


def get_transport():
    return oslo_messaging.get_notification_transport(CONF)


class Notifier(object):
    """Uses a notification strategy to send out messages about events."""

    def __init__(self):
        publisher_id = CONF.default_publisher_id
        self._transport = get_transport()
        self._notifier = oslo_messaging.Notifier(self._transport,
                                                 publisher_id=publisher_id)

    def warn(self, event_type, payload):
        self._notifier.warn({}, event_type, payload)

    def info(self, event_type, payload):
        self._notifier.info({}, event_type, payload)

    def error(self, event_type, payload):
        self._notifier.error({}, event_type, payload)


def _get_notification_group(notification):
    return notification.split('.', 1)[0]


def _is_notification_enabled(notification):
    disabled_notifications = CONF.disabled_notifications
    notification_group = _get_notification_group(notification)

    notifications = (notification, notification_group)
    for disabled_notification in disabled_notifications:
        if disabled_notification in notifications:
            return False

    return True


def _send_notification(notify, notification_type, payload):
    if _is_notification_enabled(notification_type):
        notify(notification_type, payload)


class NotificationBase(object):
    def get_payload(self, obj):
        return {}

    def send_notification(self, notification_id, obj, extra_payload=None):
        payload = self.get_payload(obj)
        if extra_payload is not None:
            payload.update(extra_payload)

        _send_notification(self.notifier.info, notification_id, payload)


@six.add_metaclass(abc.ABCMeta)
class NotificationProxy(NotificationBase):
    def __init__(self, repo, context, notifier):
        self.repo = repo
        self.context = context
        self.notifier = notifier

        super_class = self.get_super_class()
        super_class.__init__(self, repo)

    @abc.abstractmethod
    def get_super_class(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class NotificationRepoProxy(NotificationBase):
    def __init__(self, repo, context, notifier):
        self.repo = repo
        self.context = context
        self.notifier = notifier
        proxy_kwargs = {'context': self.context, 'notifier': self.notifier}

        proxy_class = self.get_proxy_class()
        super_class = self.get_super_class()
        super_class.__init__(self, repo, proxy_class, proxy_kwargs)

    @abc.abstractmethod
    def get_super_class(self):
        pass

    @abc.abstractmethod
    def get_proxy_class(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class NotificationFactoryProxy(object):
    def __init__(self, factory, context, notifier):
        kwargs = {'context': context, 'notifier': notifier}

        proxy_class = self.get_proxy_class()
        super_class = self.get_super_class()
        super_class.__init__(self, factory, proxy_class, kwargs)

    @abc.abstractmethod
    def get_super_class(self):
        pass

    @abc.abstractmethod
    def get_proxy_class(self):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# based on full set of fields payload version 1.0
NODE_MAPPING = {
    'dynamic': False,
    'properties': {
        # "id" field does not present in ironic API resource, this is a
        # copy of "uuid"
        'id': {'type': 'string', 'index': 'not_analyzed'},
        'uuid': {'type': 'string', 'index': 'not_analyzed'},
        'name': {
            'type': 'string',
            'fields': {
                'raw': {'type': 'string', 'index': 'not_analyzed'}
            }
        },
        'chassis_uuid': {'type': 'string', 'index': 'not_analyzed'},
        'instance_uuid': {'type': 'string', 'index': 'not_analyzed'},
        'driver': {'type': 'string', 'index': 'not_analyzed'},
        'driver_info': {'type': 'object', 'dynamic': True, 'properties': {}},
        'clean_step': {'type': 'object', 'dynamic': True, 'properties': {}},
        'instance_info': {'type': 'object', 'dynamic': True, 'properties': {}},
        # mapped "properties" field
        'node_properties': {
            'type': 'object', 'dynamic': True, 'properties': {}
        },
        'power_state': {'type': 'string', 'index': 'not_analyzed'},
        'target_power_state': {'type': 'string', 'index': 'not_analyzed'},
        'provision_state': {'type': 'string', 'index': 'not_analyzed'},
        'target_provision_state': {'type': 'string', 'index': 'not_analyzed'},
        'provision_updated_at': {'type': 'date'},
        'maintenance': {'type': 'boolean'},
        'maintenance_reason': {'type': 'string'},
        'console_enabled': {'type': 'boolean'},
        'last_error': {'type': 'string'},
        'resource_class': {'type': 'string', 'index': 'not_analyzed'},
        'inspection_started_at': {'type': 'date'},
        'inspection_finished_at': {'type': 'date'},
        'extra': {'type': 'object', 'dynamic': True, 'properties': {}},
        'network_interface': {'type': 'string', 'index': 'not_analyzed'},
        'created_at': {'type': 'date'},
        'updated_at': {'type': 'date'}
    }
}

PORT_MAPPING = {
    'dynamic': False,
    'properties': {
        # "id" field does not present in ironic API resource, this is
        # copy of "uuid"
        'id': {'type': 'string', 'index': 'not_analyzed'},
        'uuid': {'type': 'string', 'index': 'not_analyzed'},
        # "name" field does not present in ironic API resource, this is a
        # copy of "uuid"
        'name': {
            'type': 'string',
            'fields': {
                'raw': {'type': 'string', 'index': 'not_analyzed'}
            }
        },
        'node_uuid': {'type': 'string', 'index': 'not_analyzed'},
        'address': {'type': 'string', 'index': 'not_analyzed'},
        'extra': {'type': 'object', 'dynamic': True, 'properties': {}},
        'local_link_connection': {
            'type': 'object',
            'properties': {
                'switch_id': {'type': 'string', 'index': 'not_analyzed'},
                'port_id': {'type': 'string', 'index': 'not_analyzed'},
                'switch_info': {'type': 'string', 'index': 'not_analyzed'}
            }
        },
        'pxe_enabled': {'type': 'boolean'},
        'created_at': {'type': 'date'},
        'updated_at': {'type': 'date'}
    }
}

CHASSIS_MAPPING = {
    'dynamic': False,
    'properties': {
        # "id" field does not present in ironic API resource, this is a
        # copy of "uuid"
        'id': {'type': 'string', 'index': 'not_analyzed'},
        'uuid': {'type': 'string', 'index': 'not_analyzed'},
        # "name" field does not present in ironic API resource, this is a
        # copy of "uuid"
        'name': {
            'type': 'string',
            'fields': {
                'raw': {'type': 'string', 'index': 'not_analyzed'}
            }
        },
        'extra': {'type': 'object', 'dynamic': True, 'properties': {}},
        'description': {'type': 'string'},
        'created_at': {'type': 'date'},
        'updated_at': {'type': 'date'}
    }
}


NODE_FIELDS = NODE_MAPPING['properties'].keys()
PORT_FIELDS = PORT_MAPPING['properties'].keys()
CHASSIS_FIELDS = PORT_MAPPING['properties'].keys()



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from oslo_utils import uuidutils

from searchlight.tests import functional


MATCH_ALL = {"query": {"match_all": {}}}

OWNER1 = uuidutils.generate_uuid()
NETWORK_TENANT_ID = '8eaac046b2c44ab99246cb0850c7f06d'


class TestSearchListenerBase(functional.FunctionalTest):

    def __init__(self, *args, **kwargs):
        super(TestSearchListenerBase, self).__init__(*args, **kwargs)

    def _send_event_to_listener(self, event, index_to_flush):
        event = copy.deepcopy(event)
        self.notification_endpoint.info(
            event['ctxt'],
            event['publisher_id'],
            event['event_type'],
            event['payload'],
            event['metadata']
        )
        self._flush_elasticsearch(index_to_flush)

    def _verify_event_processing(self, event, count=1, owner=None):
        if not owner:
            payload = event['payload']
            # Try several keys in the payload to extract an owner
            owner = payload.get(
                'owner',
                payload.get('project_id',
                            payload.get('tenant_id')))
            if not owner:
                raise Exception("No project/owner info found in payload")

        response, json_content = self._search_request(
            MATCH_ALL,
            owner, role="admin")
        self.assertEqual(count, json_content['hits']['total'])
        return json_content

    def _verify_result(self, event, verification_keys, result_json,
                       inner_key=None):
        expected = event['payload']
        if inner_key:
            expected = expected[inner_key]

        result = result_json['hits']['hits'][0]['_source']
        for key in verification_keys:
            self.assertEqual(expected[key], result[key])



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import time

from searchlight.common import utils
import searchlight.elasticsearch.plugins.swift as swift_plugin
from searchlight.elasticsearch.plugins.swift import\
    containers as containers_plugin
import searchlight.tests.utils as test_utils


now_epoch_time = time.time()
now_utc = datetime.datetime.fromtimestamp(now_epoch_time).\
    strftime('%Y-%m-%dT%H:%M:%SZ')


USER1 = "27f4d76b-be62-4e4e-aa33bb11cc55"

ACCOUNT_ID1 = "488ac936-663e-4e5c-537d-986021b32c4b"
ACCOUNT_ID2 = "7554da43-6443-acdf-deac-3425223cdada"
ACCOUNT_ID3 = "30754354-ca43-124b-12b5-789234bcdefa'"

AUTH_PREFIX = "AUTH_"

CONTAINER1 = "Container1"
CONTAINER2 = "Container2"
CONTAINER_ID1 = ACCOUNT_ID1 + swift_plugin.ID_SEP + CONTAINER1
CONTAINER_ID2 = ACCOUNT_ID2 + swift_plugin.ID_SEP + CONTAINER2

X_CONTAINER_META_KEY1 = 'x-container-meta-key1'
X_CONTAINER_META_VALUE1 = 'x-container-meta-value1'

X_CONTAINER_META_KEY2 = 'x-container-meta-key2'
X_CONTAINER_META_VALUE2 = 'x-container-meta-value2'

TENANT1 = "15b9a454cee34dbe9933ad575a0a6930"
TENANT2 = "a7ba963f71bb43818f631febbc9df8e6"
DOMAIN_ID = "default"

DATETIME = datetime.datetime(2016, 2, 21, 4, 41, 33, 325314)
DATE1 = utils.isotime(DATETIME)


def _container_fixture(container_id, container_name, account,
                       account_id, read_acl, **kwargs):
    fixture = {
        "id": container_id,
        "name": container_name,
        "account": account,
        "account_id": account_id,
        "x-container-read": read_acl,
        'x-timestamp': now_epoch_time
    }
    fixture.update(kwargs)
    return fixture


def _notification_container_fixture(account_id, **kwargs):
    metadata = kwargs.pop('meta', {})
    notification = {
        'account': account_id,
        'project_name': None,
        'container': None,
        'updated_at': DATE1,
        'project_domain_name': None,
        'x-trans-id': None,
        'project_id': None,
        'project_domain_id': None
    }
    for k, v in kwargs.items():
        if k in notification:
            notification[k] = v
    for k, v in metadata.items():
        notification[k] = v
    return notification


class TestSwiftContainerPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSwiftContainerPlugin, self).setUp()
        self.plugin = containers_plugin.ContainerIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        self.container1 = _container_fixture(
            container_id=CONTAINER_ID1, container_name=CONTAINER1,
            account='test-account1', account_id=ACCOUNT_ID1, read_acl=None)
        self.container2 = _container_fixture(
            container_id=CONTAINER_ID2, container_name=CONTAINER2,
            account='test-account2', account_id=ACCOUNT_ID1,
            read_acl=ACCOUNT_ID2 + ":*",
            **{X_CONTAINER_META_KEY1: X_CONTAINER_META_VALUE1}
        )
        self.container3 = _container_fixture(
            container_id=CONTAINER_ID1, container_name=CONTAINER1,
            account='test-account3', account_id=ACCOUNT_ID2,
            read_acl=ACCOUNT_ID3 + ":*" + USER1,
            **{X_CONTAINER_META_KEY1: X_CONTAINER_META_VALUE1,
               X_CONTAINER_META_KEY2: X_CONTAINER_META_VALUE2})
        self.containers = [self.container1, self.container2, self.container3]

    def test_index_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Swift::Container',
                         self.plugin.get_document_type())

    def test_admin_only_fields(self):
        admin_only_fields = self.plugin.admin_only_fields
        self.assertEqual(['x-container-write',
                          'x-container-sync-key',
                          'x-container-sync-to',
                          'x-container-meta-temp-url-key',
                          'x-container-meta-temp-url-key-2'],
                         admin_only_fields)

    def test_serialize(self):
        serialized = self.plugin.serialize(self.container1)

        self.assertEqual(CONTAINER_ID1, serialized['id'])
        self.assertEqual(CONTAINER1, serialized['name'])
        self.assertEqual(ACCOUNT_ID1, serialized['account_id'])
        self.assertEqual('test-account1', serialized['account'])

        serialized = self.plugin.serialize(self.container2)
        self.assertEqual(X_CONTAINER_META_VALUE1,
                         serialized[X_CONTAINER_META_KEY1])
        self.assertEqual(ACCOUNT_ID2 + ":*", serialized['x-container-read'])

        serialized = self.plugin.serialize(self.container3)
        self.assertEqual(X_CONTAINER_META_VALUE1,
                         serialized[X_CONTAINER_META_KEY1])
        self.assertEqual(X_CONTAINER_META_VALUE2,
                         serialized[X_CONTAINER_META_KEY2])

    def test_swift_account_notification_serialize(self):
        notification = _notification_container_fixture(
            swift_plugin.AUTH_PREFIX + ACCOUNT_ID1,
            container=CONTAINER1,
            project_name='admin',
            project_domain_id='default',
            project_id=ACCOUNT_ID1,
            updated_at=DATE1,
        )

        expected = {
            'id': CONTAINER_ID1,
            'name': CONTAINER1,
            'account': 'admin',
            'account_id': ACCOUNT_ID1,
            'updated_at': DATE1,
            'x-container-read': None
        }

        serialized = swift_plugin.serialize_swift_container_notification(
            notification)
        self.assertEqual(expected, serialized)



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.neutron import notification_handlers
from searchlight.elasticsearch.plugins.neutron import serialize_router
from searchlight.elasticsearch.plugins import openstack_clients


class RouterIndex(base.IndexBase):
    NotificationHandlerCls = notification_handlers.RouterHandler

    ADMIN_ONLY_FIELDS = ['distributed', 'ha']

    @classmethod
    def get_document_type(self):
        return resource_types.NEUTRON_ROUTER

    def get_mapping(self):
        return {
            'dynamic': 'false',
            'properties': {
                'admin_state_up': {'type': 'boolean'},
                'availability_zone_hints': {'type': 'string',
                                            'index': 'not_analyzed'},
                'availability_zones': {'type': 'string',
                                       'index': 'not_analyzed'},
                # Routers don't have created_at
                'description': {'type': 'string'},
                'distributed': {'type': 'boolean'},
                'external_gateway_info': {
                    'type': 'nested',
                    'properties': {
                        'enable_snat': {'type': 'boolean'},
                        'external_fixed_ips': {
                            'type': 'nested',
                            # TODO(sjmc7) Check we can deal with arbitrary
                            # levels of nesting with facets
                            'properties': {
                                'ip_address': {'type': 'string',
                                               'index': 'not_analyzed'},
                                'subnet_id': {'type': 'string',
                                              'index': 'not_analyzed'},
                            }
                        },
                        'network_id': {'type': 'string',
                                       'index': 'not_analyzed'}
                    }
                },
                'ha': {'type': 'boolean'},
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'project_id': {'type': 'string', 'index': 'not_analyzed'},
                # TODO(sjmc7) Decide whether to keep these;
                # they seem like trouble
                'routes': {
                    'type': 'nested',
                    'properties': {
                        'destination': {'type': 'string',
                                        'index': 'not_analyzed'},
                        'nexthop': {'type': 'string', 'index': 'not_analyzed'},
                        'action': {'type': 'string', 'index': 'not_analyzed'},
                        'source': {'type': 'string',
                                   'index': 'not_analyzed'}
                    }
                },
                'status': {'type': 'string', 'index': 'not_analyzed'},
                'tenant_id': {'type': 'string', 'index': 'not_analyzed'},
                'updated_at': {'type': 'date'}
            },
            "_meta": {
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "external_gateway_info.external_fixed_ips.subnet_id": {
                    "resource_type": resource_types.NEUTRON_SUBNET
                },
                "external_gateway_info.network_id": {
                    "resource_type": resource_types.NEUTRON_NETWORK
                }
            }
        }

    @property
    def admin_only_fields(self):
        from_conf = super(RouterIndex, self).admin_only_fields
        return from_conf + RouterIndex.ADMIN_ONLY_FIELDS

    @property
    def facets_with_options(self):
        return ('admin_state_up', 'availability_zones', 'status',
                'distributed', 'ha', 'external_gateway_info.enable_snat')

    @property
    def facets_excluded(self):
        return {'tenant_id': True, 'distributed': True, 'ha': True,
                'project_id': True}

    @property
    def resource_allowed_policy_target(self):
        # Neutron only supports policy for individual routers
        return None

    @property
    def service_type(self):
        return 'network'

    def _get_rbac_field_filters(self, request_context):
        return [
            {'term': {'tenant_id': request_context.owner}}
        ]

    def get_objects(self):
        neutron_client = openstack_clients.get_neutronclient()
        for router in neutron_client.list_routers()['routers']:
            yield router

    def serialize(self, router):
        return serialize_router(router)
